@@ -46,10 +46,21 @@ import type Parser from "./index";
 const loopLabel = { kind: "loop" } as const,
   switchLabel = { kind: "switch" } as const;
 
-const FUNC_NO_FLAGS = 0b000,
-  FUNC_STATEMENT = 0b001,
-  FUNC_HANGING_STATEMENT = 0b010,
-  FUNC_NULLABLE_ID = 0b100;
+export const enum ParseFunctionFlag {
+  Expression = 0b0000,
+  Declaration = 0b0001,
+  HangingDeclaration = 0b0010,
+  NullableId = 0b0100,
+  Async = 0b1000,
+}
+
+export const enum ParseStatementFlag {
+  StatementOnly = 0b0000,
+  AllowImportExport = 0b0001,
+  AllowDeclaration = 0b0010,
+  AllowFunctionDeclaration = 0b0100,
+  AllowLabeledFunction = 0b1000,
+}
 
 const loneSurrogate = /[\uD800-\uDFFF]/u;
 
@@ -62,8 +73,6 @@ const keywordRelationalOperator = /in(?:stanceof)?/y;
  * tt.templateNonTail => tt.backquote/tt.braceR + tt.template + tt.dollarBraceL
  * For performance reasons this routine mutates `tokens`, it is okay
  * here since we execute `parseTopLevel` once for every file.
- * @param {*} tokens
- * @returns
  */
 function babel7CompatTokens(tokens: (Token | N.Comment)[], input: string) {
   for (let i = 0; i < tokens.length; i++) {
@@ -237,14 +246,8 @@ export default abstract class StatementParser extends ExpressionParser {
     return finishedProgram;
   }
 
-  // TODO
-
   /**
    * cast a Statement to a Directive. This method mutates input statement.
-   *
-   * @param {N.Statement} stmt
-   * @returns {N.Directive}
-   * @memberof StatementParser
    */
   stmtToDirective(stmt: N.Statement): N.Directive {
     const directive = stmt as any;
@@ -277,42 +280,16 @@ export default abstract class StatementParser extends ExpressionParser {
     return this.finishNode(node, "InterpreterDirective");
   }
 
-  isLet(context?: string | null): boolean {
+  isLet(): boolean {
     if (!this.isContextual(tt._let)) {
       return false;
     }
-    return this.hasFollowingIdentifier(context);
+    return this.hasFollowingBindingAtom();
   }
 
-  /**
-   * Assuming we have seen a contextual `let`, check if it starts a variable declaration
-   so that `left` should be interpreted as a `let` keyword.
-   *
-   * @param {?string} context When `context` is non nullish, it will return early and _skip_ checking
-                              if the next token after `let` is `{` or a keyword relational operator
-   * @returns {boolean}
-   * @memberof StatementParser
-   */
-  hasFollowingIdentifier(context?: string | null): boolean {
-    const next = this.nextTokenStart();
-    const nextCh = this.codePointAtPos(next);
-    // For ambiguous cases, determine if a LexicalDeclaration (or only a
-    // Statement) is allowed here. If context is not empty then only a Statement
-    // is allowed. However, `let [` is an explicit negative lookahead for
-    // ExpressionStatement, so special-case it first.
-    // Also, `let \` is never valid as an expression so this must be a keyword.
-    if (
-      nextCh === charCodes.backslash ||
-      nextCh === charCodes.leftSquareBracket
-    ) {
-      return true;
-    }
-    if (context) return false;
-
-    if (nextCh === charCodes.leftCurlyBrace) return true;
-
-    if (isIdentifierStart(nextCh)) {
-      keywordRelationalOperator.lastIndex = next;
+  chStartsBindingIdentifier(ch: number, pos: number) {
+    if (isIdentifierStart(ch)) {
+      keywordRelationalOperator.lastIndex = pos;
       if (keywordRelationalOperator.test(this.input)) {
         // We have seen `in` or `instanceof` so far, now check if the identfier
         // ends here
@@ -322,8 +299,40 @@ export default abstract class StatementParser extends ExpressionParser {
         }
       }
       return true;
+    } else if (ch === charCodes.backslash) {
+      return true;
+    } else {
+      return false;
     }
-    return false;
+  }
+
+  chStartsBindingPattern(ch: number) {
+    return (
+      ch === charCodes.leftSquareBracket || ch === charCodes.leftCurlyBrace
+    );
+  }
+
+  /**
+   * Assuming we have seen a contextual `let` and declaration is allowed, check if it
+   * starts a variable declaration so that it should be interpreted as a keyword.
+   */
+  hasFollowingBindingAtom(): boolean {
+    const next = this.nextTokenStart();
+    const nextCh = this.codePointAtPos(next);
+    return (
+      this.chStartsBindingPattern(nextCh) ||
+      this.chStartsBindingIdentifier(nextCh, next)
+    );
+  }
+
+  /**
+   * Assuming we have seen a contextual `using` and declaration is allowed, check if it
+   * starts a variable declaration so that it should be interpreted as a keyword.
+   */
+  hasFollowingBindingIdentifier(): boolean {
+    const next = this.nextTokenStart();
+    const nextCh = this.codePointAtPos(next);
+    return this.chStartsBindingIdentifier(nextCh, next);
   }
 
   startsUsingForOf(): boolean {
@@ -337,6 +346,35 @@ export default abstract class StatementParser extends ExpressionParser {
     }
   }
 
+  // https://tc39.es/ecma262/#prod-ModuleItem
+  parseModuleItem(this: Parser) {
+    return this.parseStatementLike(
+      ParseStatementFlag.AllowImportExport |
+        ParseStatementFlag.AllowDeclaration |
+        ParseStatementFlag.AllowFunctionDeclaration |
+        ParseStatementFlag.AllowLabeledFunction,
+    );
+  }
+
+  // https://tc39.es/ecma262/#prod-StatementListItem
+  parseStatementListItem(this: Parser) {
+    return this.parseStatementLike(
+      ParseStatementFlag.AllowDeclaration |
+        ParseStatementFlag.AllowFunctionDeclaration |
+        ParseStatementFlag.AllowLabeledFunction,
+    );
+  }
+
+  parseStatementOrFunctionDeclaration(
+    this: Parser,
+    disallowLabeledFunction: boolean,
+  ) {
+    return this.parseStatementLike(
+      ParseStatementFlag.AllowFunctionDeclaration |
+        (disallowLabeledFunction ? 0 : ParseStatementFlag.AllowLabeledFunction),
+    );
+  }
+
   // Parse a single statement.
   //
   // If expecting a statement and finding a slash operator, parse a
@@ -344,29 +382,42 @@ export default abstract class StatementParser extends ExpressionParser {
   // `if (foo) /blah/.exec(foo)`, where looking at the previous token
   // does not help.
   // https://tc39.es/ecma262/#prod-Statement
+  parseStatement(this: Parser) {
+    return this.parseStatementLike(ParseStatementFlag.StatementOnly);
+  }
+
   // ImportDeclaration and ExportDeclaration are also handled here so we can throw recoverable errors
   // when they are not at the top level
-  parseStatement(
+  parseStatementLike(
     this: Parser,
-    context?: string | null,
-    topLevel?: boolean,
-  ): N.Statement {
+    flags: ParseStatementFlag,
+  ):
+    | N.Statement
+    | N.Declaration
+    | N.ImportDeclaration
+    | N.ExportDefaultDeclaration
+    | N.ExportNamedDeclaration
+    | N.ExportAllDeclaration {
     let decorators: N.Decorator[] | null = null;
 
     if (this.match(tt.at)) {
       decorators = this.parseDecorators(true);
     }
-    return this.parseStatementContent(context, topLevel, decorators);
+    return this.parseStatementContent(flags, decorators);
   }
 
   parseStatementContent(
     this: Parser,
-    context?: string | null,
-    topLevel?: boolean | null,
+    flags: ParseStatementFlag,
     decorators?: N.Decorator[] | null,
   ): N.Statement {
     const starttype = this.state.type;
     const node = this.startNode();
+    const allowDeclaration = !!(flags & ParseStatementFlag.AllowDeclaration);
+    const allowFunctionDeclaration = !!(
+      flags & ParseStatementFlag.AllowFunctionDeclaration
+    );
+    const topLevel = flags & ParseStatementFlag.AllowImportExport;
 
     // Most types of statements are recognized by the keyword they
     // start with. Many are trivial to parse, some require a bit of
@@ -380,26 +431,25 @@ export default abstract class StatementParser extends ExpressionParser {
       case tt._debugger:
         return this.parseDebuggerStatement(node as Undone<N.DebuggerStatement>);
       case tt._do:
-        return this.parseDoStatement(node as Undone<N.DoWhileStatement>);
+        return this.parseDoWhileStatement(node as Undone<N.DoWhileStatement>);
       case tt._for:
         return this.parseForStatement(node as Undone<N.ForStatement>);
       case tt._function:
         if (this.lookaheadCharCode() === charCodes.dot) break;
-        if (context) {
+        if (!allowDeclaration) {
           if (this.state.strict) {
             this.raise(Errors.StrictFunction, { at: this.state.startLoc });
-          } else if (context !== "if" && context !== "label") {
+          } else if (!allowFunctionDeclaration) {
             this.raise(Errors.SloppyFunction, { at: this.state.startLoc });
           }
         }
         return this.parseFunctionStatement(
           node as Undone<N.FunctionDeclaration>,
           false,
-          !context,
+          !allowDeclaration && allowFunctionDeclaration,
         );
-
       case tt._class:
-        if (context) this.unexpected();
+        if (!allowDeclaration) this.unexpected();
         return this.parseClass(
           this.maybeTakeDecorators(
             decorators,
@@ -421,31 +471,56 @@ export default abstract class StatementParser extends ExpressionParser {
 
       case tt._using:
         // using [no LineTerminator here] BindingList[+Using]
-        if (this.hasFollowingLineBreak()) {
+        if (
+          this.hasFollowingLineBreak() ||
+          this.state.containsEsc ||
+          !this.hasFollowingBindingIdentifier()
+        ) {
           break;
         }
-      // fall through
-      case tt._let:
-        if (this.state.containsEsc || !this.hasFollowingIdentifier(context)) {
-          break;
-        }
-      // fall through
-      case tt._const:
-      case tt._var: {
-        const kind = this.state.value;
-        if (kind === "using") {
-          this.expectPlugin("explicitResourceManagement");
-          if (!this.scope.inModule && this.scope.inTopLevel) {
-            this.raise(Errors.UnexpectedUsingDeclaration, {
-              at: this.state.startLoc,
-            });
-          }
-        }
-        if (context && kind !== "var") {
+        this.expectPlugin("explicitResourceManagement");
+        if (!this.scope.inModule && this.scope.inTopLevel) {
+          this.raise(Errors.UnexpectedUsingDeclaration, {
+            at: this.state.startLoc,
+          });
+        } else if (!allowDeclaration) {
           this.raise(Errors.UnexpectedLexicalDeclaration, {
             at: this.state.startLoc,
           });
         }
+        return this.parseVarStatement(
+          node as Undone<N.VariableDeclaration>,
+          "using",
+        );
+      case tt._let: {
+        if (this.state.containsEsc) {
+          break;
+        }
+        // `let [` is an explicit negative lookahead for
+        // ExpressionStatement, so special-case it first.
+        const next = this.nextTokenStart();
+        const nextCh = this.codePointAtPos(next);
+        if (nextCh !== charCodes.leftSquareBracket) {
+          if (!allowDeclaration && this.hasFollowingLineBreak()) break;
+          if (
+            !this.chStartsBindingIdentifier(nextCh, next) &&
+            nextCh !== charCodes.leftCurlyBrace
+          ) {
+            break;
+          }
+        }
+      }
+      // fall through
+      case tt._const: {
+        if (!allowDeclaration) {
+          this.raise(Errors.UnexpectedLexicalDeclaration, {
+            at: this.state.startLoc,
+          });
+        }
+      }
+      // fall through
+      case tt._var: {
+        const kind = this.state.value;
         return this.parseVarStatement(
           node as Undone<N.VariableDeclaration>,
           kind,
@@ -516,16 +591,16 @@ export default abstract class StatementParser extends ExpressionParser {
 
       default: {
         if (this.isAsyncFunction()) {
-          if (context) {
+          if (!allowDeclaration) {
             this.raise(Errors.AsyncFunctionInSingleStatementContext, {
               at: this.state.startLoc,
             });
           }
-          this.next();
+          this.next(); // eat 'async'
           return this.parseFunctionStatement(
             node as Undone<N.FunctionDeclaration>,
             true,
-            !context,
+            !allowDeclaration && allowFunctionDeclaration,
           );
         }
       }
@@ -549,7 +624,7 @@ export default abstract class StatementParser extends ExpressionParser {
         maybeName,
         // @ts-expect-error migrate to Babel types
         expr,
-        context,
+        flags,
       );
     } else {
       return this.parseExpressionStatement(
@@ -749,7 +824,8 @@ export default abstract class StatementParser extends ExpressionParser {
     return val;
   }
 
-  parseDoStatement(
+  // https://tc39.es/ecma262/#prod-DoWhileStatement
+  parseDoWhileStatement(
     this: Parser,
     node: Undone<N.DoWhileStatement>,
   ): N.DoWhileStatement {
@@ -763,7 +839,7 @@ export default abstract class StatementParser extends ExpressionParser {
       // outside of the loop body.
       this.withSmartMixTopicForbiddingContext(() =>
         // Parse the loop body's body.
-        this.parseStatement("do"),
+        this.parseStatement(),
       );
 
     this.state.labels.pop();
@@ -808,9 +884,9 @@ export default abstract class StatementParser extends ExpressionParser {
     const startsWithUsing =
       this.isContextual(tt._using) && !this.hasFollowingLineBreak();
     const isLetOrUsing =
-      (startsWithLet && this.hasFollowingIdentifier()) ||
+      (startsWithLet && this.hasFollowingBindingAtom()) ||
       (startsWithUsing &&
-        this.hasFollowingIdentifier() &&
+        this.hasFollowingBindingIdentifier() &&
         this.startsUsingForOf());
     if (this.match(tt._var) || this.match(tt._const) || isLetOrUsing) {
       const initNode = this.startNode<N.VariableDeclaration>();
@@ -881,25 +957,35 @@ export default abstract class StatementParser extends ExpressionParser {
     return this.parseFor(node as Undone<N.ForStatement>, init);
   }
 
+  // https://tc39.es/ecma262/#prod-HoistableDeclaration
   parseFunctionStatement(
     this: Parser,
     node: Undone<N.FunctionDeclaration>,
-    isAsync?: boolean,
-    declarationPosition?: boolean,
+    isAsync: boolean,
+    isHangingDeclaration: boolean,
   ): N.FunctionDeclaration {
-    this.next();
+    this.next(); // eat 'function'
     return this.parseFunction(
       node,
-      FUNC_STATEMENT | (declarationPosition ? 0 : FUNC_HANGING_STATEMENT),
-      isAsync,
+      ParseFunctionFlag.Declaration |
+        (isHangingDeclaration ? ParseFunctionFlag.HangingDeclaration : 0) |
+        (isAsync ? ParseFunctionFlag.Async : 0),
     );
   }
 
+  // https://tc39.es/ecma262/#prod-IfStatement
   parseIfStatement(this: Parser, node: Undone<N.IfStatement>) {
     this.next();
     node.test = this.parseHeaderExpression();
-    node.consequent = this.parseStatement("if");
-    node.alternate = this.eat(tt._else) ? this.parseStatement("if") : null;
+    // Annex B.3.3
+    // https://tc39.es/ecma262/#sec-functiondeclarations-in-ifstatement-statement-clauses
+    node.consequent = this.parseStatementOrFunctionDeclaration(
+      // https://tc39.es/ecma262/#sec-if-statement-static-semantics-early-errors
+      true,
+    );
+    node.alternate = this.eat(tt._else)
+      ? this.parseStatementOrFunctionDeclaration(true)
+      : null;
     return this.finishNode(node, "IfStatement");
   }
 
@@ -924,6 +1010,7 @@ export default abstract class StatementParser extends ExpressionParser {
     return this.finishNode(node, "ReturnStatement");
   }
 
+  // https://tc39.es/ecma262/#prod-SwitchStatement
   parseSwitchStatement(this: Parser, node: Undone<N.SwitchStatement>) {
     this.next();
     node.discriminant = this.parseHeaderExpression();
@@ -959,7 +1046,7 @@ export default abstract class StatementParser extends ExpressionParser {
         this.expect(tt.colon);
       } else {
         if (cur) {
-          cur.consequent.push(this.parseStatement(null));
+          cur.consequent.push(this.parseStatementListItem());
         } else {
           this.unexpected();
         }
@@ -1039,6 +1126,8 @@ export default abstract class StatementParser extends ExpressionParser {
     return this.finishNode(node, "TryStatement");
   }
 
+  // https://tc39.es/ecma262/#prod-VariableStatement
+  // https://tc39.es/ecma262/#prod-LexicalDeclaration
   parseVarStatement(
     this: Parser,
     node: Undone<N.VariableDeclaration>,
@@ -1051,6 +1140,7 @@ export default abstract class StatementParser extends ExpressionParser {
     return this.finishNode(node, "VariableDeclaration");
   }
 
+  // https://tc39.es/ecma262/#prod-WhileStatement
   parseWhileStatement(
     this: Parser,
     node: Undone<N.WhileStatement>,
@@ -1066,7 +1156,7 @@ export default abstract class StatementParser extends ExpressionParser {
       // They are permitted in test expressions, outside of the loop body.
       this.withSmartMixTopicForbiddingContext(() =>
         // Parse loop body.
-        this.parseStatement("while"),
+        this.parseStatement(),
       );
 
     this.state.labels.pop();
@@ -1092,7 +1182,7 @@ export default abstract class StatementParser extends ExpressionParser {
       // part of the outer context, outside of the with statement's body.
       this.withSmartMixTopicForbiddingContext(() =>
         // Parse the statement body.
-        this.parseStatement("with"),
+        this.parseStatement(),
       );
 
     return this.finishNode(node, "WithStatement");
@@ -1103,12 +1193,13 @@ export default abstract class StatementParser extends ExpressionParser {
     return this.finishNode(node, "EmptyStatement");
   }
 
+  // https://tc39.es/ecma262/#prod-LabelledStatement
   parseLabeledStatement(
     this: Parser,
     node: Undone<N.LabeledStatement>,
     maybeName: string,
     expr: N.Identifier,
-    context?: string | null,
+    flags: ParseStatementFlag,
   ): N.LabeledStatement {
     for (const label of this.state.labels) {
       if (label.name === maybeName) {
@@ -1139,13 +1230,11 @@ export default abstract class StatementParser extends ExpressionParser {
       kind: kind,
       statementStart: this.state.start,
     });
-    node.body = this.parseStatement(
-      context
-        ? context.indexOf("label") === -1
-          ? context + "label"
-          : context
-        : "label",
-    );
+    // https://tc39.es/ecma262/#prod-LabelledItem
+    node.body =
+      flags & ParseStatementFlag.AllowLabeledFunction
+        ? this.parseStatementOrFunctionDeclaration(false)
+        : this.parseStatement();
 
     this.state.labels.pop();
     node.label = expr;
@@ -1238,7 +1327,9 @@ export default abstract class StatementParser extends ExpressionParser {
     let parsedNonDirective = false;
 
     while (!this.match(end)) {
-      const stmt = this.parseStatement(null, topLevel);
+      const stmt = topLevel
+        ? this.parseModuleItem()
+        : this.parseStatementListItem();
 
       if (directives && !parsedNonDirective) {
         if (this.isValidDirective(stmt)) {
@@ -1296,7 +1387,7 @@ export default abstract class StatementParser extends ExpressionParser {
       // outside of the loop body.
       this.withSmartMixTopicForbiddingContext(() =>
         // Parse the loop body.
-        this.parseStatement("for"),
+        this.parseStatement(),
       );
 
     this.scope.exit();
@@ -1357,7 +1448,7 @@ export default abstract class StatementParser extends ExpressionParser {
       // They are permitted in test expressions, outside of the loop body.
       this.withSmartMixTopicForbiddingContext(() =>
         // Parse loop body.
-        this.parseStatement("for"),
+        this.parseStatement(),
       );
 
     this.scope.exit();
@@ -1417,9 +1508,6 @@ export default abstract class StatementParser extends ExpressionParser {
     kind: "var" | "let" | "const" | "using",
   ): void {
     const id = this.parseBindingAtom();
-    if (kind === "using" && id.type !== "Identifier") {
-      this.raise(Errors.UsingDeclarationHasBindingPattern, { at: id });
-    }
     this.checkLVal(id, {
       in: { type: "VariableDeclarator" },
       binding: kind === "var" ? BIND_VAR : BIND_LEXICAL,
@@ -1427,29 +1515,40 @@ export default abstract class StatementParser extends ExpressionParser {
     decl.id = id;
   }
 
-  // Parse a function declaration or literal (depending on the
-  // `isStatement` parameter).
+  // https://tc39.es/ecma262/#prod-AsyncFunctionExpression
+  parseAsyncFunctionExpression(
+    this: Parser,
+    node: Undone<N.FunctionExpression>,
+  ): N.FunctionExpression {
+    return this.parseFunction(node, ParseFunctionFlag.Async);
+  }
+
+  // Parse a function declaration or expression (depending on the
+  // ParseFunctionFlag.Declaration flag).
 
   parseFunction<T extends N.NormalFunction>(
     this: Parser,
     node: Undone<T>,
-    statement: number = FUNC_NO_FLAGS,
-    isAsync: boolean = false,
+    flags: ParseFunctionFlag = ParseFunctionFlag.Expression,
   ): T {
-    const isStatement = statement & FUNC_STATEMENT;
-    const isHangingStatement = statement & FUNC_HANGING_STATEMENT;
-    const requireId = !!isStatement && !(statement & FUNC_NULLABLE_ID);
+    const hangingDeclaration = flags & ParseFunctionFlag.HangingDeclaration;
+    const isDeclaration = !!(flags & ParseFunctionFlag.Declaration);
+    const requireId = isDeclaration && !(flags & ParseFunctionFlag.NullableId);
+    const isAsync = !!(flags & ParseFunctionFlag.Async);
 
     this.initFunction(node, isAsync);
 
-    if (this.match(tt.star) && isHangingStatement) {
-      this.raise(Errors.GeneratorInSingleStatementContext, {
-        at: this.state.startLoc,
-      });
+    if (this.match(tt.star)) {
+      if (hangingDeclaration) {
+        this.raise(Errors.GeneratorInSingleStatementContext, {
+          at: this.state.startLoc,
+        });
+      }
+      this.next(); // eat *
+      node.generator = true;
     }
-    node.generator = this.eat(tt.star);
 
-    if (isStatement) {
+    if (isDeclaration) {
       node.id = this.parseFunctionId(requireId);
     }
 
@@ -1458,7 +1557,7 @@ export default abstract class StatementParser extends ExpressionParser {
     this.scope.enter(SCOPE_FUNCTION);
     this.prodParam.enter(functionFlags(isAsync, node.generator));
 
-    if (!isStatement) {
+    if (!isDeclaration) {
       node.id = this.parseFunctionId();
     }
 
@@ -1471,14 +1570,14 @@ export default abstract class StatementParser extends ExpressionParser {
       // Parse the function body.
       this.parseFunctionBodyAndFinish(
         node,
-        isStatement ? "FunctionDeclaration" : "FunctionExpression",
+        isDeclaration ? "FunctionDeclaration" : "FunctionExpression",
       );
     });
 
     this.prodParam.exit();
     this.scope.exit();
 
-    if (isStatement && !isHangingStatement) {
+    if (isDeclaration && !hangingDeclaration) {
       // We need to register this _after_ parsing the function body
       // because of TypeScript body-less function declarations,
       // which shouldn't be added to the scope.
@@ -2339,18 +2438,20 @@ export default abstract class StatementParser extends ExpressionParser {
   parseExportDefaultExpression(this: Parser): N.Expression | N.Declaration {
     const expr = this.startNode();
 
-    const isAsync = this.isAsyncFunction();
-
-    if (this.match(tt._function) || isAsync) {
+    if (this.match(tt._function)) {
       this.next();
-      if (isAsync) {
-        this.next();
-      }
-
       return this.parseFunction(
-        expr as Undone<N.FunctionExpression>,
-        FUNC_STATEMENT | FUNC_NULLABLE_ID,
-        isAsync,
+        expr as Undone<N.FunctionDeclaration>,
+        ParseFunctionFlag.Declaration | ParseFunctionFlag.NullableId,
+      );
+    } else if (this.isAsyncFunction()) {
+      this.next(); // eat 'async'
+      this.next(); // eat 'function'
+      return this.parseFunction(
+        expr as Undone<N.FunctionDeclaration>,
+        ParseFunctionFlag.Declaration |
+          ParseFunctionFlag.NullableId |
+          ParseFunctionFlag.Async,
       );
     }
 
@@ -2386,6 +2487,7 @@ export default abstract class StatementParser extends ExpressionParser {
     return res;
   }
 
+  // https://tc39.es/ecma262/#prod-ExportDeclaration
   parseExportDeclaration(
     this: Parser,
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
@@ -2399,7 +2501,7 @@ export default abstract class StatementParser extends ExpressionParser {
       );
       return node;
     }
-    return this.parseStatement(null) as N.Declaration;
+    return this.parseStatementListItem() as N.Declaration;
   }
 
   isExportDefaultSpecifier(): boolean {
@@ -2747,8 +2849,9 @@ export default abstract class StatementParser extends ExpressionParser {
     let isImportReflection = false;
     if (this.isContextual(tt._module)) {
       const lookahead = this.lookahead();
-      if (tokenIsIdentifier(lookahead.type)) {
-        if (lookahead.type !== tt._from) {
+      const nextType = lookahead.type;
+      if (tokenIsIdentifier(nextType)) {
+        if (nextType !== tt._from) {
           // import module x
           isImportReflection = true;
         } else {
@@ -2760,9 +2863,10 @@ export default abstract class StatementParser extends ExpressionParser {
             isImportReflection = true;
           }
         }
-      } else {
+      } else if (nextType !== tt.comma) {
         // import module { x } ...
-        // This is invalid, we will continue parsing and throw
+        // import module "foo"
+        // They are invalid, we will continue parsing and throw
         // a recoverable error later
         isImportReflection = true;
       }
@@ -2864,9 +2968,7 @@ export default abstract class StatementParser extends ExpressionParser {
   /**
    * parse assert entries
    *
-   * @see {@link https://tc39.es/proposal-import-assertions/#prod-AssertEntries |AssertEntries}
-   * @returns {N.ImportAttribute[]}
-   * @memberof StatementParser
+   * @see {@link https://tc39.es/proposal-import-assertions/#prod-AssertEntries AssertEntries}
    */
   parseAssertEntries(): N.ImportAttribute[] {
     const attrs = [];
@@ -2913,8 +3015,6 @@ export default abstract class StatementParser extends ExpressionParser {
   /**
    * parse module attributes
    * @deprecated It will be removed in Babel 8
-   * @returns
-   * @memberof StatementParser
    */
   maybeParseModuleAttributes() {
     if (this.match(tt._with) && !this.hasPrecedingLineBreak()) {
